@@ -39,7 +39,9 @@ let finalize_stmt stmt =
   Sqlite3.finalize stmt |> fail_if_not_ok
 
 let final_step stmt =
-  Sqlite3.step stmt |> expect Sqlite3.Rc.DONE
+  Utils.with_retry (fun () ->
+    Sqlite3.step stmt |> expect Sqlite3.Rc.DONE)
+    "final_step"
 (* END Helpers *)
 
 (* Query helpers *)
@@ -102,7 +104,7 @@ let select_all_rows stmt bind_parameters row_to_data =
 
 (* Prepare SQL *)
 let prepare_begin_tran_stmt db =
-  Sqlite3.prepare db "BEGIN TRANSACTION;"
+  Sqlite3.prepare db "BEGIN IMMEDIATE TRANSACTION;"
 
 let prepare_commit_tran_stmt db =
   Sqlite3.prepare db "COMMIT TRANSACTION;"
@@ -195,6 +197,15 @@ struct
     in
       Sqlite3.prepare db sql
 
+  let prepare_update_state_stmt db =
+    let sql =
+      "UPDATE resource \
+       SET \
+         state = :state \
+       WHERE id = :id;"
+    in
+      Sqlite3.prepare db sql
+
   let prepare_update_all_change_ids db =
     let sql =
       "UPDATE resource \
@@ -228,7 +239,8 @@ struct
          state = 'ToDownload', \
          change_id = :change_id \
        WHERE id = :id \
-         AND state <> 'ToUpload';"
+         AND state <> 'ToUpload' \
+         AND state <> 'Uploading';"
     in
       Sqlite3.prepare db sql
 
@@ -398,19 +410,9 @@ let open_db cache =
   db
 
 let close_db db =
-  let rec try_close n =
-    if n > 4 then begin
-      let thread_id = Utils.get_thread_id () in
-      Utils.log_message "Thread id=%d: Error: cannot close db\n%!" thread_id;
-      Printf.eprintf "Error: cannot close sqlite db.\n\
-        Please restart the program.\n%!";
-      exit 1
-    end else if not (Sqlite3.db_close db) then begin
-      Unix.sleep 1;
-      try_close (succ n)
-    end
-  in
-  try_close 0
+  Utils.with_retry (fun () ->
+    if not (Sqlite3.db_close db) then raise (Failure "close_db"))
+    "close_db"
 
 let with_db cache f =
   let db = open_db cache in
@@ -438,19 +440,25 @@ struct
     type t =
         InSync
       | ToDownload
+      | Downloading
       | ToUpload
+      | Uploading
       | NotFound
 
     let to_string = function
         InSync -> "InSync"
       | ToDownload -> "ToDownload"
+      | Downloading -> "Downloading"
       | ToUpload -> "ToUpload"
+      | Uploading -> "Uploading"
       | NotFound -> "NotFound"
 
     let of_string = function
         "InSync" -> InSync
       | "ToDownload" -> ToDownload
+      | "Downloading" -> Downloading
       | "ToUpload" -> ToUpload
+      | "Uploading" -> Uploading
       | "NotFound" -> NotFound
       | s -> failwith ("Resource state unexpected: " ^ s)
 
@@ -660,7 +668,7 @@ struct
          insert_stmt db)
 
   let update_resource cache resource =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_update_stmt db in
          bind_resource_parameters stmt resource;
@@ -668,20 +676,29 @@ struct
          final_step stmt;
          finalize_stmt stmt)
 
+  let update_resource_state cache state id =
+    with_transaction cache
+      (fun db ->
+         let stmt = ResourceStmts.prepare_update_state_stmt db in
+         bind_text stmt ":state" (Some (State.to_string state));
+         bind_int stmt ":id" (Some id);
+         final_step stmt;
+         finalize_stmt stmt)
+
   let _delete_resource stmt id =
     reset_stmt stmt;
     bind_int stmt ":id" (Some id);
     final_step stmt
 
   let delete_resource cache resource =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_delete_stmt db in
          _delete_resource stmt resource.id;
          finalize_stmt stmt)
 
   let delete_not_found_resource_with_path cache path =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_delete_not_found_with_path_stmt db in
          reset_stmt stmt;
@@ -737,7 +754,7 @@ struct
          finalize_stmt stmt)
 
   let invalidate_trash_bin cache change_id =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_invalidate_trash_bin_stmt db in
          reset_stmt stmt;
@@ -759,7 +776,7 @@ struct
          finalize_stmt stmt)
 
   let delete_all_with_parent_path cache parent_path trashed =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_delete_all_with_parent_path db in
          reset_stmt stmt;
@@ -769,7 +786,7 @@ struct
          finalize_stmt stmt)
 
   let trash_all_with_parent_path cache parent_path =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_trash_all_with_parent_path db in
          reset_stmt stmt;
@@ -803,7 +820,7 @@ struct
     }
 
   let select_resource cache prepare bind =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = prepare db in
          let result =
@@ -824,7 +841,7 @@ struct
       (fun stmt -> bind_text stmt ":remote_id" (Some remote_id))
 
   let select_resources_with_parent_path cache parent_path trashed =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_select_with_parent_path_stmt db in
          let results =
@@ -837,7 +854,7 @@ struct
          results)
 
   let select_resources_order_by_last_update cache =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = ResourceStmts.prepare_select_order_by_last_update db in
          let results =
@@ -892,6 +909,27 @@ struct
         Some mime_type -> get_format_from_mime_type mime_type config
       | _ -> "html"
 
+  let get_icon_from_mime_type mime_type config =
+    match mime_type with
+        "application/vnd.google-apps.document" ->
+          config.Config.document_icon
+      | "application/vnd.google-apps.drawing" ->
+          config.Config.drawing_icon
+      | "application/vnd.google-apps.form" ->
+          config.Config.form_icon
+      | "application/vnd.google-apps.presentation" ->
+          config.Config.presentation_icon
+      | "application/vnd.google-apps.spreadsheet" ->
+          config.Config.spreadsheet_icon
+      | "application/vnd.google-apps.map" ->
+          config.Config.map_icon
+      | _ -> ""
+
+  let get_icon resource config =
+    match resource.mime_type with
+        Some mime_type -> get_icon_from_mime_type mime_type config
+      | _ -> "html"
+
   let mime_type_of_format fmt =
     match fmt with
         "doc"
@@ -986,7 +1024,7 @@ struct
     final_step stmt
 
   let insert_metadata cache resource =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = MetadataStmts.prepare_insert_stmt db in
          save_metadata stmt resource;
@@ -1006,7 +1044,7 @@ struct
     }
 
   let select_metadata cache =
-    with_db cache
+    with_transaction cache
       (fun db ->
          let stmt = MetadataStmts.prepare_select_stmt db in
          let result =
@@ -1040,11 +1078,14 @@ let delete_files_from_cache cache resources =
   List.fold_left
     (fun total_size resource ->
        let content_path = get_content_path cache resource in
-       Utils.log_message "Removing file (%s: resource %Ld) from cache...%!"
+       Utils.log_with_header
+         "BEGIN: Removing file (%s: resource %Ld) from cache\n%!"
          content_path resource.Resource.id;
        let size = remove_file content_path in
        let new_size = Int64.add total_size size in
-       Utils.log_message "done\n%!";
+       Utils.log_with_header
+         "END: Removing file (%s: resource %Ld) from cache\n%!"
+         content_path resource.Resource.id;
        new_size)
     0L
     resources
@@ -1054,7 +1095,7 @@ let setup_db cache =
   with_db cache
     (fun db ->
       wrap_exec_not_null_no_headers db
-        "BEGIN TRANSACTION; \
+        "BEGIN IMMEDIATE TRANSACTION; \
          CREATE TABLE IF NOT EXISTS resource ( \
             id INTEGER PRIMARY KEY, \
             etag TEXT NULL, \
@@ -1107,7 +1148,7 @@ let clean_up_cache cache =
          try
            Sys.remove (Filename.concat cache.cache_dir file)
          with e ->
-           Utils.log_message "Error removing file %s: %s\n%!"
+           Utils.log_with_header "Error removing file %s: %s\n%!"
              file (Printexc.to_string e))
       (Sys.readdir cache.cache_dir)
   end
